# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


class SingletonObject(object):
    """Use to create a singleton"""

    def __new__(cls):
        self = "__self__"
        if not hasattr(cls, self):
            instance = object.__new__(cls)
            instance.__init__()
            setattr(cls, self, instance)
        return getattr(cls, self)

    def __repr__(self):
        return str(self) + '(' + str(id(self)) + ')'

    def __str__(self):
        return self.__class__.__name__


"""
module containing GenericFactory class
"""


class GenericFactory(object):
    """
    implementation of generic __factory used for object creation
    """

    # instance = None
    #
    # @staticmethod
    # def get_instance():
    #     if GenericFactory.instance is None:
    #         GenericFactory.instance = GenericFactory()
    #
    #     return GenericFactory.instance

    def register(self, object_id, creator):
        """
        register object creator for given id

        :param object object_id : object_id. It must be immutable.
        :param object creator : object creator
        :return bool :
        """

        res = False
        if object_id not in self.__factory_dict:
            self.__factory_dict[object_id] = creator
            res = True
            if not self.__default_value:
                self.__default_value = creator

        return res

    def create_object(self, object_id=None, *create_params):
        """
        create object based on the given id and given parameters

        :param object object_id : id of constructed object (e.g. class name)
        :param iterable create_params : parameters required for creation (if any)
        :return object :
        """

        if not object_id:
            object_id = self.__default_value
        res = self.__internal_create_object(object_id)
        if len(create_params) > 0:
            res = res(*create_params)

        return res

    def __init__(self):
        """
        initializer
        """

        #: dictionary to keep track of the objects
        self.__factory_dict = dict()
        self.__default_value = None

    def keys(self):
        """
        returns keys in the __factory

        @return list :
        """

        return self.__factory_dict.keys()

    def __internal_create_object(self, object_id):
        """
        returns object creator functor based on object_id

        :param object object_id : immutable object id
        :return object :
        """

        return self.__factory_dict[object_id]



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


import inspect


class LinkedObject(object):
    """ links from linked_obj to (obj, attribute) with obj.attribute = linked_obj """
    __link = dict()

    @property
    def _name(self):
        return str(self)

    @classmethod
    def _get_links(cls):
        mro = inspect.getmro(cls)
        for m in mro:
            attr = '_' + m.__name__ + '__link'
            if hasattr(cls, attr):
                return getattr(cls, attr)
        raise TypeError

    def __repr__(self):
        return self.__class__.__name__ + '.' + str(self) + '(' + str(id(self)) + ')'

    def __str__(self):
        return str(self.__class__.__name__)

    def __setattr__(self, item, value):
        if hasattr(self, item):
            current = getattr(self, item)
            if isinstance(current, LinkedObject):
                current.remove_link(self, item)
        super(LinkedObject, self).__setattr__(item, value)
        if isinstance(value, LinkedObject):
            value.register_link(self, item)

    def register_link(self, obj, attr=None):
        """
        creates link from obj.attr to self
        :param obj: object to register link to
        :param attr: attribute name to register link to
        """
        name = self._name
        l = self.__class__._get_links()
        if name not in l:
            l[name] = set()
        v = (obj, attr)
        if v not in l:
            l[name].add(v)
        return self

    def remove_link(self, obj, attr=None):
        """
        removes link from obj.attr
        """
        name = self._name
        l = self.__class__._get_links()
        v = (None, obj) if attr is None else (obj, attr)
        if v in l[name]:
            l[name].remove(v)
        if not l[name]:
            l.pop(name)
        return self

    def update_link(self):
        """
        redirects all links to self (the new linked object)
        """
        name = self._name
        l = self.__class__._get_links()
        to_be_changed = list()
        if name in l:
            for o, a in l[name]:
                if self is not getattr(o, a, None):
                    to_be_changed.append((o,a))
        for o, a in to_be_changed:
            setattr(o, a, self)

        return self



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


class IndexedList(list):
    def __init__(self, keys=list(), values=None):
        # fixme imput no or empty list (just headers)
        values = keys if values is None else values
        keys = [str(v) for v in values] if keys is values else keys

        if isinstance(values, IndexedList):
            keys = values.keys()
            values = values.values()
        if len(values) is not len(keys):
            raise IndexError, '%s requires exactly one key for each item' % self.__class__.__name__
        if len(set(keys)) is not len(keys):
            raise ValueError, 'In %s keys must be unique' % self.__class__.__name__
        self._keys = dict(zip(keys, range(len(values))))
        super(IndexedList, self).__init__(values)

    def __add__(self, other):
        assert isinstance(other, IndexedList)
        keys = self.keys() + other.keys()
        values = self.values() + other.values()
        return self.__class__(values, keys)

    def __iadd__(self, other):
        assert isinstance(other, IndexedList)
        keys = self.keys() + other.keys()
        values = self.values() + other.values()
        return self.__class__(values, keys)

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            s = key.start if key.start else 0
            e = key.stop if key.stop else len(self)
            self.__setslice__(s, e, value)
        else:
            index = self._keys[key] if key in self._keys else key
            super(IndexedList, self).__setitem__(index, value)

    def __getitem__(self, item):
        if isinstance(item, slice):
            s = item.start if item.start else 0
            e = item.stop if item.stop else len(self)
            return self.__getslice__(s, e)
        else:
            index = self._keys[item] if item in self._keys else item
            return super(IndexedList, self).__getitem__(index)

    def __delitem__(self, key):
        if isinstance(key, slice):
            s = key.start if key.start else 0
            e = key.stop if key.stop else len(self)
            self.__delslice__(s, e)
        else:
            index = self._keys[key] if key in self._keys else key
            super(IndexedList, self).__delitem__(index)

    def __setslice__(self, i, j, sequence):
        i_index = self._keys[i] if i in self._keys else i
        j_index = self._keys[j] + 1 if j in self._keys else j
        super(IndexedList, self).__setslice__(i_index, j_index, sequence)

    def __getslice__(self, i, j):
        i_index = self._keys[i] if i in self._keys else i
        i_index = len(self) if i is None else i_index
        j_index = self._keys[j] + 1 if j in self._keys else j
        j_index = len(self) if j is None else j_index
        return super(IndexedList, self).__getslice__(i_index, j_index)

    def __delslice__(self, i, j):
        i_index = self._keys[i] if i in self._keys else i
        j_index = self._keys[j] + 1 if j in self._keys else j
        super(IndexedList, self).__delslice__(i_index, j_index)

    def insert(self, item, key, value=None):
        index = self._keys[item] if item in self._keys else item
        value = key if value is None else value
        key = str(value) if key is value else key

        # validate key
        if key in self._keys:
            raise KeyError, 'Key %s already given.' % str(item)
        super(IndexedList, self).insert(index, value)
        # rebuild _keys
        keys = self.keys()
        keys.insert(index, key)
        self._keys = dict(zip(keys, range(len(self))))

    def append(self, key, value=None):
        value = key if value is None else value
        key = str(value) if key is value else key
        # validate key
        if key in self._keys:
            raise KeyError, 'Key %s already given.' % str(key)
        self._keys[key] = len(self)
        super(IndexedList, self).append(value)

    def extend(self, keys, values=None):
        values = keys if values is None else values
        keys = [str(v) for v in values] if keys is values else keys

        if isinstance(values, IndexedList):
            keys = values.keys()
            values = values.values()

        for k in keys:
            if k in self._keys:
                raise KeyError, 'Key %s already given.' % str(k)
            self._keys[k] = len(self._keys)
        super(IndexedList, self).extend(values)

    def keys(self):
        l = (lambda x, y: self._keys[x] - self._keys[y])
        return sorted(self._keys.keys(), l)

    def values(self):
        return [self[k] for k in self.keys()]

    def items(self):
        return [(k, self[k]) for k in self.keys()]


class DataRange(IndexedList):
    """ DataRange class """

    def __init__(self, iterable=None, value_types=(float, int, str, type(None)), none_alias=(None, ' ', '', None)):
        self._value_types = value_types
        self._none_alias = none_alias

        # covert dict into nested list
        if isinstance(iterable, dict):
            iterable = DataRange.__dict_to_nested_list(iterable)

        # replace None alias by None
        none_alias = none_alias if isinstance(none_alias, (tuple, list)) else [none_alias]
        if iterable:
            f = (lambda x: None if x in none_alias else x)
            iterable = [[f(c) for c in r] for r in iterable]

        # slice nested list iterable into (column headers, row headers, data)
        ch, rh, values = DataRange.__slice_nested_list(iterable)
        # todo validate column header entries for ambiguity (no int < len(col_headers))
        # todo validate row header entries for ambiguity (no int < len(row_headers))
        # todo validate iterable (only admissible value types)

        super(DataRange, self).__init__(rh, [IndexedList(ch, v) for v in values])

    @staticmethod
    def __dict_to_nested_list(iterable):
        item_list = [iterable.keys()]
        for i in xrange(max(map(len, iterable.values()))):
            item_list.append([iterable[k][i] for k in item_list[0]])
        return item_list

    @staticmethod
    def __slice_nested_list(iterable):
        if not iterable:
            return list(), list(), list()

        # extract column headers from first row
        col_header = iterable.pop(0)
        if len(set(col_header)) is not len(col_header):
            raise ValueError, 'All column header entries must be unique.'

        # extract row headers if given
        if col_header.count(None):
            i = col_header.index(None)
            col_header.pop(i)
            row_header = [row.pop(i) for row in iterable]
        else:
            row_header = range(len(iterable))
        if len(set(row_header)) is not len(row_header):
            raise ValueError, 'All row header entries must be unique.'

        return col_header, row_header, iterable

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % str(id(self))

    def __str__(self):
        return self.__class__.__name__ + '(%s)' % str(self.to_serializable())

    def __iter__(self):
        return super(DataRange, self).__iter__()

    def __setitem__(self, key, value):
        # todo validate value (only admissible value types)

        if isinstance(key, slice):
            return self.__setslice__(key.start, key.stop)

        if key in self.row_keys() or type(key) is int:
            return super(DataRange, self).__setitem__(key, value)

        if len(key) == 2:
            r, c = key
            return super(DataRange, self).__getitem__(r).__setitem__(c, value)

        raise KeyError, 'Key %s not found.' % repr(key)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__getslice__(item.start, item.stop)

        if item in self.row_keys() or type(item) is int:
            return super(DataRange, self).__getitem__(item)

        if len(item) == 2:
            r, c = item
            return super(DataRange, self).__getitem__(r).__getitem__(c)

        raise KeyError, 'Key %s not found.' % repr(item)

    def __delitem__(self, key):
        raise NotImplementedError

    def __setslice__(self, i, j, sequence):
        raise NotImplementedError

    def __getslice__(self, i, j):
        if isinstance(i, (tuple, list)):
            row_start, col_start = i
            row_stop, col_stop = j
            row_start = row_start if row_start else 0
            row_stop = row_stop if row_stop is not None else len(self.row_keys())
            col_start = col_start if col_start else 0
            col_stop = col_stop if col_stop is not None else len(self.col_keys())
            return_list = [d.__getslice__(col_start, col_stop) for d in
                           super(DataRange, self).__getslice__(row_start, row_stop)]
        else:
            i = 0 if i is None else i
            j = len(self.row_keys()) if j is None else j
            return_list = super(DataRange, self).__getslice__(i, j)
        # return self.__class__(return_list, value_types=self._value_types, none_alias=self._none_alias)
        return return_list

    def __delslice__(self, i, j):
        raise NotImplementedError

    def insert(self, item, key, value=None):
        raise NotImplementedError

    def pop(self, index=None):
        raise NotImplementedError

    def append(self, key, value=None):
        if value is None:
            value = [None] * len(self.col_keys())
        if not isinstance(value, list) and not len(value) == len(self.col_keys()):
            raise ValueError
        if isinstance(value, IndexedList):
            if not value.keys() == self.col_keys():
                raise KeyError
        else:
            value = IndexedList(self.col_keys(), value)
        super(DataRange, self).append(key, value)

    def extend(self, keys, values=None):
        raise NotImplementedError

    def transpose(self):
        return self.__class__(zip(*self._total), value_types=self._value_types, none_alias=self._none_alias)

    def row_keys(self):
        return list(super(DataRange, self).keys())

    def col_keys(self):
        return list(super(DataRange, self).__getitem__(0).keys())

    def keys(self):
        keys = list()
        for row_key in self.row_keys():
            for col_key in self.col_keys():
                keys.append((row_key, col_key))
        return keys

    def values(self):
        return [self[k] for k in self.keys()]

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    @property
    def _total(self):
        if not self:
            return [[]]
        ret = [[None] + self.col_keys()]
        for r in self.row_keys():
            l = [r]
            for v in self[r]:
                l.append(v)
            ret.append(l)
        return ret

    def to_serializable(self, level=0):
        ret = list()
        for r in self._total:
            l = list()
            for v in r:
                v = v if not hasattr(v, 'to_serializable') else v.to_serializable(level + 1)
                v = self._none_alias[0] if isinstance(v, type(None)) else v
                v = v if isinstance(v, (float, int, type(None))) else str(v)
                l.append(v)
            ret.append(l)
        return ret

    def row(self, item):
        return self[item]

    def col(self, item):
        return [self[r][item] for r in self.row_keys()]



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


class CachedObject(type):
    _cache = dict()
    _link = dict()

    def __call__(cls, name=''):
        name = 'Core' + cls.__name__  if name is '' else name
        if name not in CachedObject._cache:
            c = cls.__new__(cls, name)
            c.__init__(name)
            CachedObject._cache[name] = c
        return CachedObject._cache[name]

    @staticmethod
    def get(name, default=None):
        return CachedObject._cache[name] if name in CachedObject._cache else default

    @staticmethod
    def all(**kwargs):
        if kwargs:
            is_type = kwargs.pop('isinstance', None)
            l = list()
            for o in CachedObject._cache:
                obj = CachedObject._cache[o]
                if all([hasattr(obj, k) for k in kwargs]):
                    if all([getattr(obj, k) == kwargs[k] for k in kwargs]):
                        l.append(obj)
            if is_type:
                for o in l:
                    if isinstance(o, is_type):
                        l.remove(o)
            return l
        else:
            return CachedObject._cache

    @staticmethod
    def flush(name=None):
        if name is None:
            CachedObject._cache = dict()
        elif name in CachedObject._cache:
            CachedObject._cache.pop(name)

    @staticmethod
    def register(name, obj, attr):
        if name not in CachedObject._link:
            CachedObject._link[name] = set()
        if (obj, attr) not in CachedObject._link[name]:
            CachedObject._link[name].add((obj, attr))

    @staticmethod
    def update(name, value):
        CachedObject._cache[name] = value
        if name in CachedObject._link:
            for o, a in CachedObject._link[name]:
                if value is not getattr(o, a):
                    setattr(o, a, value)


# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


import inspect


class FactoryType(type):

    def __call__(cls, *args, **kwargs):
        name = str(args[0]) if args else cls.__name__
        name = kwargs['name'] if 'name' in kwargs else name
        instance = cls._get(name)
        if instance is None:
            instance = object.__new__(cls, *args, **kwargs)
            instance.__init__(*args, **kwargs)
        return instance


class FactoryObject(object):
    """ Objects identified by name """
    __factory = dict()

    __metaclass__ = FactoryType

    @property
    def _name(self):
        return str(self)

    @classmethod
    def _get_factory(cls):
        mro = inspect.getmro(cls)
        for m in mro:
            attr = '_' + m.__name__ + '__factory'
            if hasattr(cls, attr):
                return getattr(cls, attr)
        raise TypeError

    def x__new__(cls, *args):
        name = str(args[0]) if args else cls.__name__
        instance = cls._get(name)
        if instance is None:
            instance = object.__new__(cls, *args)
            instance.__init__(*args)
        return instance

    def x__init__(self, *args):
        super(FactoryObject, self).__init__()

    def __repr__(self):
        return str(self) + '(' + str(id(self)) + ')'

    def __str__(self):
        return str(self.__class__.__name__)

    def to_serializable(self, level=0):
        return self._name

    def register(self, *names):
        if not names:
            names = (self._name,)
        for name in names:
            self.__class__._set(str(name), self)
        return self

    def remove(self):
        for k in self.__class__._keys(self):
            self.__class__._del(k)
        return self

    @classmethod
    def _set(cls, key, value):
        f = cls._get_factory()
        f[key] = value

    @classmethod
    def _get(cls, key, default=None):
        f = cls._get_factory()
        if key is not None and key in f:
            return f[key]
        else:
            return default

    @classmethod
    def _del(cls, key):
        f = cls._get_factory()
        if key is not None and key in f:
            f.pop(key)

    @classmethod
    def _keys(cls, value=None):
        f = cls._get_factory()
        if value is None:
            return f.keys()
        return [k for k, v in f.items() if value == v]

    @classmethod
    def from_serializable(cls, item):
        return cls(item)

    @classmethod
    def items(cls):
        return cls._get_factory().items()

    @classmethod
    def keys(cls):
        return cls._keys()

    @classmethod
    def values(cls):
        return cls._get_factory().values()


class ObjectList(list):

    def __init__(self, iterable=None, object_type=FactoryObject):
        if not issubclass(object_type, FactoryObject):
            raise TypeError, 'Required object type of ObjectList items must be subtype of %s ' % FactoryObject.__name__
        self._object_type = object_type
        if iterable is None:
            super(ObjectList, self).__init__()
        else:
            i = list()
            for x in iterable:
                if type(x) is not str:
                    self.__validate(x)
                else:
                    x = self._object_type(x)
                i.append(x)
            iterable = i
            for x in iterable:
                self.__validate(x)
            super(ObjectList, self).__init__(iterable)

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self) + ', %s)' % self._object_type.__name__

    def __str__(self):
        return '[' +  ', '.join(repr(x) for x in self) + ']'

    def __validate(self, x):
        if not isinstance(x, self._object_type):
            raise TypeError, 'All items in this ObjectList must be of subtype of %s ' %self._object_type.__name__

    def __iter__(self):
        return super(ObjectList, self).__iter__()

    def __setitem__(self, key, value):
        value = self._object_type(value)
        self.__validate(value)
        super(ObjectList, self).__setitem__(key, value)

    def __contains__(self, item):
        if isinstance(item, self._object_type):
            return super(ObjectList, self).__contains__(item)
        return item in [str(f) for f in self]

    def index(self, item, start=None, stop=None):
        if start is None or stop is None:
            if isinstance(item, self._object_type):
                return super(ObjectList, self).index(item)
            elif type(item) is str:
                return [str(f) for f in self].index(item)
        return [f.to_serializable() for f in self].index(item, start, stop)

    def __getitem__(self, item):
        if isinstance(item, self._object_type):
            return [o for o in self if o == item][0]
        elif type(item) is str:
            return [o for o in self if str(o) == item][0]
        return super(ObjectList, self).__getitem__(item)

    def get(self, item, default=None):
        return self[item] if item in self else default

    def __add__(self, other):
        other = [self._object_type(value) for value in other]
        for value in other:
            self.__validate(value)
        return ObjectList(super(ObjectList, self).__add__(other), self._object_type)

    def __iadd__(self, other):
        other = [self._object_type(value) for value in other]
        for value in other:
            self.__validate(value)
        return self.__class__(super(ObjectList, self).__iadd__(other), self._object_type)

    def __setslice__(self, i, j, iterable):
        iterable = [self._object_type(value) for value in iterable]
        for value in iterable:
            self.__validate(value)
        super(ObjectList, self).__setslice__(i, j, iterable)

    def append(self, value):
        value = self._object_type(value)
        self.__validate(value)
        super(ObjectList, self).append(value)

    def insert(self, index, value):
        value = self._object_type(value)
        self.__validate(value)
        super(ObjectList, self).insert(index, value)

    def extend(self, iterable):
        iterable = [self._object_type(value) for value in iterable]
        for value in iterable:
            self.__validate(value)
        super(ObjectList, self).extend(iterable)

    def to_serializable(self, level=0):
        return [x.to_serializable(level + 1) for x in self]

    @classmethod
    def from_serializable(cls, item):
        return cls(item)

    def __reduce__(self):
        return self.__class__, (list(self), self._object_type)



"""
  module containing StringSerialization mixin
"""


class StringSerialization(object):
    """
     implementation of mixin for/to string serialization

    :version:
    :author:
    """

    # definition for pylint (real FactoryObject should be defined in subclass)
    Factory = None

    def to_string(self):
        """
        returns __name__ attribute of the instance used  used by persistence framework
        for object persistence

        @return string :
        @author
        """
        return type(self).__name__

    @classmethod
    def from_string(cls, name):
        """
        returns day count convention instance based on the name

        @param string name : object_id used for construction from cls.FactoryObject \
        e.g. Act360, Act365, D30360
        @return object : e.g businessdate.DayCount
        @author
        """
        return cls.Factory.create_object(name)

    def __repr__(self):
        return self.to_string()

    def __str__(self):
        return type(super(StringSerialization,self)).__class__.__name__ + '(' + self.to_string() + ')'



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


class DataRange(dict):
    def __init__(self, iterable=None, value_types=(float, int, str, type(None)), none_alias=(None, ' ', '', None),
                 **kwargs):
        self._value_types = value_types
        self._none_alias = none_alias
        self._col_keys = list()
        self._row_keys = list()

        # convert dict into nested list
        if isinstance(iterable, dict):
            iterable = DataRange.__dict_to_nested_list(iterable)

        # replace None alias by None
        none_alias = none_alias if isinstance(none_alias, (tuple, list)) else [none_alias]
        if iterable:
            f = (lambda x: None if x in none_alias else x)
            iterable = [[f(c) for c in r] for r in iterable]

        # slice nested list iterable into (column headers, row headers, data)
        col_keys, row_keys, values = DataRange.__slice_nested_list(iterable)

        # validate iterable (only admissible value types)
        for row_value in values:
            for value in row_value:
                self._validate_value(value)

        # validate column header entries for ambiguity (no int)
        # if any(isinstance(col_key, int) for col_key in col_keys):
        #    raise KeyError('Column keys must not be of type int.')

        # validate row header entries for ambiguity (no int < len(row_headers))
        # if any(isinstance(row_key, int) for row_key in row_keys):
        #    raise KeyError('If row keys are of type int, values not must between 0 and %i.' % len(row_keys))

        if not len(col_keys) == len(set(col_keys)):
            raise KeyError('Col keys must be unique.')
        if not len(row_keys) == len(set(row_keys)):
            raise KeyError('Row keys must be unique.')

        self._col_keys = col_keys
        self._row_keys = row_keys

        iterable = list()
        for row_key, row_values in zip(row_keys, values):
            for col_key, value in zip(col_keys, row_values):
                iterable.append(((row_key, col_key), value))

        super(DataRange, self).__init__(iterable, **kwargs)

    @staticmethod
    def __dict_to_nested_list(iterable):
        item_list = [iterable.keys()]
        for i in xrange(max(map(len, iterable.values()))):
            item_list.append([iterable[k][i] for k in item_list[0]])
        return item_list

    @staticmethod
    def __slice_nested_list(iterable):
        if not iterable:
            return list(), list(), list()

        # extract column headers from first row
        col_header = iterable.pop(0)
        if not len(set(col_header)) == len(col_header):
            raise ValueError, 'All column header entries must be unique.'

        # extract row headers if given
        if col_header.count(None):
            i = col_header.index(None)
            col_header.pop(i)
            row_header = [row.pop(i) for row in iterable]
        else:
            row_header = range(len(iterable))
            if not len(set(row_header)) == len(row_header):
                raise ValueError, 'All row header entries must be unique.'

        return col_header, row_header, iterable

    def _validate_value(self, value):
        if not isinstance(value, self._value_types):
            s = ', '.join([str(t) for t in self._value_types])
            t = type(value)
            raise TypeError, \
                'All properties of item in this AttributeList must have type ' \
                'of either one of %s but not %s.' % (s, t)
        return True

    def _validate_key(self, key):
        if not isinstance(key, (list, tuple)) or not len(key) == 2:
            raise KeyError('Key of %s must be (row_key, col_key) tuple.' % self.__class__.__name__)

        r, c = key
        if r not in self._row_keys or c not in self._col_keys:
            s = self.__class__.__name__, str(self._row_keys), str(self._col_keys)
            raise KeyError('Key of %s must be (row_key, col_key) tuple with \n '
                           'row_key in %s \n and col_key in %s' % s)
        return True

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % str(id(self))

    def __str__(self):
        return self.__class__.__name__ + '(%s)' % str(self.to_serializable())

    def __len__(self):
        return len(self._row_keys)

    def __eq__(self, other):
        assert isinstance(other, DataRange)
        return self.total_list == other.total_list

    def __contains__(self, key):
        try:
            self._validate_key(key)
        except KeyError:
            return False
        return True

    def __copy__(self):
        return self.__class__(self.total_list)

    def __deepcopy__(self, memo):
        return self.__class__(self.to_serializable())

    def __setitem__(self, key, value):
        self._validate_key(key)
        self._validate_value(value)
        super(DataRange, self).__setitem__(key, value)

    def __getitem__(self, key):
        # redirect slice
        if isinstance(key, slice):
            return self.__getslice__(key.start, key.stop)

        # gather which row or col is requested
        # if key in self._row_keys or type(key) is int:
        #     return self.row(key)
        # elif key in self._col_keys:
        #     return self.col(key)

        if not isinstance(key, (list, tuple)) or not len(key) == 2:
            raise KeyError('Key of %s must be row key, col_key or (row key, col key) tuple.' % self.__class__.__name__)

        r, c = key
        # try to pic key if int given
        r = self._row_keys[r] if type(r) is int else r
        c = self._col_keys[c] if type(c) is int else c
        if r not in self._row_keys or c not in self._col_keys:
            s = self.__class__.__name__, str(self._row_keys), str(self._col_keys)
            raise KeyError('Key of %s must be (row_key, col_key) tuple with \n row_key in %s \n and col_key in %s' % s)

        return super(DataRange, self).get((r, c), None)

    def __delitem__(self, key):
        raise NotImplementedError

    def __setslice__(self, i, j, sequence):
        raise NotImplementedError

    def __getslice__(self, i, j):
        if not isinstance(i, (tuple, list)):
            l = len(self._col_keys)
            return self[(i, 0):(j, l)]

        ri, ci = i
        rj, cj = j
        ri = self._row_index(ri)
        rj = self._row_index(rj) + 1 if rj in self._row_keys else rj
        ci = self._col_index(ci)
        cj = self._col_index(cj) + 1 if cj in self._col_keys else cj
        row_keys = self._row_keys[ri:rj]
        col_keys = self._col_keys[ci:cj]
        ret = list()
        for r in row_keys:
            row = [self[r,c] for c in col_keys]
            ret.append(row)
        return ret

    def __delslice__(self, i, j):
        raise NotImplementedError

    def _row_index(self, key):
        return self._row_keys.index(key) if self._row_keys.count(key) else key

    def _col_index(self, key):
        return self._col_keys.index(key) if self._col_keys.count(key) else key

    def row_append(self, key, value):
        """ append new row """
        if key in self._row_keys:
            raise KeyError('Key %s already exists in row keys.' % key)
        if not isinstance(value, (tuple, list)):
            value = [value] * len(self._col_keys)
        if not len(value) == len(self._col_keys):
            raise ValueError('Length of data to set does not meet expected row length of %i' % len(self._col_keys))
        self._row_keys.append(key)
        for c, v in zip(self._col_keys, value):
            self[key, c] = v

    def col_append(self, key, value):
        if key in self._col_keys:
            raise KeyError('Key %s already exists col keys.' % key)
        if not isinstance(value, (tuple, list)):
            value = [value] * len(self._row_keys)
        if not len(value) == len(self._row_keys):
            raise ValueError('Length of data to set does not meet expected col length of %i' % len(self._row_keys))
        self._col_keys.append(key)
        for r, v in zip(self._row_keys, value):
            self[r, key] = v

    def row_keys(self):
        return self._row_keys

    def col_keys(self):
        return self._col_keys

    def row(self, item):
        r = self._row_keys[item] if type(item) == int else item
        return [self.get((r, c)) for c in self._col_keys]

    def col(self, item):
        c = self._col_keys[item] if type(item) == int else item
        return [self.get((r, c)) for r in self._row_keys]

    @property
    def item_list(self):
        return [self.row(r) for r in self.row_keys()]

    @property
    def total_list(self):
        if not self:
            return [[]]
        head = [[None] + self.col_keys()]
        body = [[r] + self.row(r) for r in self.row_keys()]
        return head + body

    def to_serializable(self, level=0):
        ret = list()
        for r in self.total_list:
            l = list()
            for v in r:
                v = v if not hasattr(v, 'to_serializable') else v.to_serializable(level + 1)
                v = self._none_alias[0] if isinstance(v, type(None)) else v
                v = v if isinstance(v, (float, int, type(None))) else str(v)
                l.append(v)
            ret.append(l)
        return ret

    def transpose(self):
        return self.__class__(zip(*self.total_list), value_types=self._value_types, none_alias=self._none_alias)

    def append(self, key, value):
        self.row_append(key, value)

    def extend(self, key, value):
        for k,v in zip(key, value):
            self.row_append(k, v)

    def insert(self, item, key, value=None):
        raise NotImplementedError

    def __reduce__(self):

        return self.__class__, (self.total_list, self._value_types, self._none_alias)




# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


from factoryobject import FactoryObject, ObjectList
from linkedobject import LinkedObject
from persistentobject import PersistentObject, AttributeList
from datarange import DataRange


class VisibleObject(FactoryObject, LinkedObject, PersistentObject):
    def __init__(self, *args, **kwargs):
        super(VisibleObject, self).__init__(*args, **kwargs)
        name = str(args[0]) if args else self.__class__.__name__
        name = kwargs['name'] if 'name' in kwargs else name
        self._name_ = name

    @property
    def _name(self):
        return self._name_

    def __repr__(self):
        return str(self) + '(' + str(id(self)) + ')'

    def __str__(self):
        return self.__class__.__name__ + '(' + self._name + ')'

    def to_serializable(self, level=0, all_properties=False):
        if level is 0:
            return PersistentObject.to_serializable(self, all_properties=all_properties)
        else:
            return FactoryObject.to_serializable(self)

    @classmethod
    def from_serializable(cls, item):
        if isinstance(item, list):
            return [o for o in VisibleAttributeList.from_serializable(item)]
        elif isinstance(item, dict):
            return PersistentObject.from_serializable(item)
        else:
            return FactoryObject.from_serializable(str(item))


class VisibleObjectList(ObjectList):
    def __init__(self, iterable=None, object_type=VisibleObject):
        super(VisibleObjectList, self).__init__(iterable, object_type)


class VisibleAttributeList(AttributeList):
    def __init__(self, iterable=None, object_type=VisibleObject,
                 value_types=(float, int, str, type(None), VisibleObject)):
        super(VisibleAttributeList, self).__init__(iterable, object_type, value_types)


class VisibleDataRange(DataRange):
    def __init__(self, iterable=None,
                 value_types=(float, int, str, type(None), VisibleObject),
                 none_alias=(None, ' ', '', None)):
        super(VisibleDataRange, self).__init__(iterable, value_types, none_alias)



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


import functools

'''
    cache keys are not case sensitive, eg actACT for ActAct
    cache returns proper capitalisation, eg ActAct for actact
    caches named objects by obj.name
    caches unnamed objects by obj.__class__.__name__

    class actact(daycount): pass

    daycount()==cache('daycount')
    daycount()==actact()
    daycount('30/360')==Thirty360()

    pricer()==

'''


def _to_string(cls):
    return str(cls.__name__)


class _CaseInsensitiveDict(dict):
    """ dict with case insensitive keys """

    def __setitem__(self, k, d):
        """ handle key case insensitive """
        return super(_CaseInsensitiveDict, self).__setitem__(k.upper(), d)

    def __contains__(self, k):
        """ handle key case insensitive """
        return super(_CaseInsensitiveDict, self).__contains__(k.upper())

    def __getitem__(self, k):
        """ handle key case insensitive """
        try:
            return super(_CaseInsensitiveDict, self). __getitem__(k.upper())
        except:
            return None

    def __delitem__(self, k):
        """ handle key case insensitive """
        return super(_CaseInsensitiveDict, self).__delitem__(k.upper())

    def pop(self, k, d=None):
        """ handle key case insensitive """
        return super(_CaseInsensitiveDict, self).pop(k.upper(), d)


class Factory(object):
    """ FactoryObject class to manage cache of named objects.
        Behaves like a dict. Is callable to factor items. """

    def __init__(self, cache=None, required_type=None, default_key=None, key_func=(lambda x: x.to_serializable())):
        self.cache = cache
        if not self.cache:
            self.cache = _CaseInsensitiveDict()
        if required_type and not isinstance(required_type, type):
            # type might be wrapper used for decorator
            required_type = required_type.__closure__[0].cell_contents
        self.type = required_type
        self.default_key = default_key
        self.key_func = key_func

    def __call__(self, key=None):
        if not key:
            key = self.default_key
        if not key:
            raise KeyError, '__factory default key not set.'
        value = self.cache[key]
        if self.type:
            assert isinstance(value, self.type)
        return value

    def __str__(self):
        if self.type:
            d = [k for k in self.cache if isinstance(self.cache[k],self.type)]
            l = max([len(k) for k in d])
            return '\n'.join([self.__class__.__name__ + '.' + self.type.__name__ + '.' \
                    + k.ljust(l)+':'+repr(self.cache[k]) for k in d])

        l = max([len(k) for k in self.cache])
        return '\n'.join([self.__class__.__name__ + '.' + k.ljust(l,)+':'+repr(self.cache[k]) for k in self.cache])

    def __setitem__(self, k, d):
        if not hasattr(d,'to_serializable'):
            setattr(d.__class__, 'to_serializable', classmethod(_to_string))
        return self.cache.__setitem__(k, d)

    def __contains__(self, k):
        return self.cache.__contains__(k)

    def __getitem__(self, k):
        return self.cache.__getitem__(k)

    def __delitem__(self, k):
        return self.cache.__delitem__(k)

    def pop(self, k, d=None):
        return self.cache.pop(k, d)


class DecoratorFactory(Factory):
    """ FactoryObject class to manage cache of named objects.
        Behaves like a dict. Is callable to factor items.
        Provides easy to use as decorators. """

    def by_auto(self, obj):
        @functools.wraps(obj)
        def factory_wrapper(*args, **kwargs):

            # retrieve key entry
            if not args:
                if self.default_key:
                    # default to self.default_key as key
                    key = self.default_key
                else:
                    # default to obj.__name__ as key
                    key = obj.__name__
            elif len(args) == 1:
                # in case of a single arg (name), take this entry as key
                key = str(args[0])
            else:
                # else store all call args
                key = str(args)

            # retrieve value entry
            if key in self.cache:
                # grab item identified by first argument
                value = self.cache[key]
            else:
                # if no item can be identified create a new one
                value = obj(args)
                # add the identified item to keys given by arguments
                self.cache[key] = value
                if obj.__name__ not in self:
                    # if no default is set do it now
                    self.cache[obj.__name__] = value

            if self.type:
                assert isinstance(value, self.type)
            return value

        return factory_wrapper

    def by_args(self, obj):
        @functools.wraps(obj)
        def factory_wrapper_by_args(*args):
            # retrieve item key
            key = str(args)
            # retrieve item
            if key not in self.cache:
                self.cache[key] = obj(args)
            value = self.cache[key]
            if self.type:
                assert isinstance(value, self.type)
            return value

        return factory_wrapper_by_args

    def by_key(self, obj):
        @functools.wraps(obj)
        def factory_wrapper_by_key(*args):
            # if no args given return obj but not cached
            if not args:
                return obj(args)
            # retrieve item key
            key = str(args[0])
            # retrieve item
            if key not in self.cache:
                self.cache[key] = obj(args)
            value = self.cache[key]
            if self.type:
                assert isinstance(value, self.type)
            return value
        return factory_wrapper_by_key

    def by_class_name(self, obj):
        @functools.wraps(obj)
        def factory_wrapper_by_class_name(*args):
            # retrieve item key
            key = obj.__name__
            # retrieve item
            if key not in self.cache:
                self.cache[key] = obj(*args)
            value = self.cache[key]
            if self.type:
                assert isinstance(value, self.type)
            return value
        return factory_wrapper_by_class_name




# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


import datetime
import getpass
import json

from decode_json import decode_dict as _decode_dict


class PersistentObject(object):
    STARTS_WITH = '_'
    ENDS_WITH = '_'
    JSON_INDENT = 2

    @property
    def is_modified(self):
        return len(self._modified_members) > 0

    @property
    def _class(self):
        return str(self.__class__.__name__)

    @property
    def _module(self):
        return str(self.__module__)

    def __init__(self, *args, **kwargs):
        super(PersistentObject, self).__init__()
        setattr(self, self.__class__.STARTS_WITH + 'class' + self.__class__.ENDS_WITH, self._class)
        setattr(self, self.__class__.STARTS_WITH + 'module' + self.__class__.ENDS_WITH, self._module)
        self._ts_create = datetime.datetime.now()
        self._user_create = getpass.getuser() if hasattr(getpass, 'getuser') else 'NoUser'
        self._ts_update = self._ts_create
        self._user_update = self._user_create
        self._modified_members = []
        self._version = 0

    def __repr__(self):
        return self.__class__.__name__ + '.' + str(self) + '(' + str(id(self)) + ')'

    def __str__(self):
        return str(self.__class__.__name__)

    @classmethod
    def _is_visible(cls, property_name):
        """ private method to check visible object property to be visible """
        if isinstance(property_name, list):
            return [cls._is_visible(p) for p in property_name]
        if property_name.startswith('__') and property_name.endswith('__'):
            return False
        return property_name.startswith(cls.STARTS_WITH) and property_name.endswith(cls.ENDS_WITH)

    @classmethod
    def _from_visible(cls, property_name):
        if isinstance(property_name, list):
            return [cls._from_visible(p) for p in property_name]
        if cls._is_visible(property_name):
            property_name = property_name[len(cls.STARTS_WITH):-len(cls.ENDS_WITH)]
            m_list = str(property_name).split('_')
            n = [l.capitalize() for l in m_list if l]
            return ''.join(n)
        return property_name

    @classmethod
    def _to_visible(cls, property_name):
        if isinstance(property_name, list):
            return [cls._to_visible(p) for p in property_name]
        if cls._is_visible(property_name):
            return property_name
        n = [property_name[0].lower()]
        for l in property_name[1:]:
            n.append('_' + l.lower()) if l.isupper() else n.append(l)
        return cls.STARTS_WITH + ''.join(n) + cls.ENDS_WITH

    @classmethod
    def _from_class(cls, class_name, module_name=None, *args, **kwargs):
        """ class method to create object of a given class """

        if module_name:
            # module = globals()[module_name]
            module = __import__(module_name)
            class_ = getattr(module, class_name)
        else:
            class_ = globals()[class_name]

        if not issubclass(class_, PersistentObject):
            t = class_.__name__, PersistentObject.__name__
            raise TypeError, 'Requested object type %s must be subtype of %s ' % t

        # workaround to mimic FactoryType to work well with FactoryObject.
        name = str(args[0]) if args else cls.__name__
        name = kwargs['name'] if 'name' in kwargs else name
        if hasattr(cls, '_get'):
            instance = cls._get(name)
            if instance:
                return instance

        instance = class_.__new__(class_, *args, **kwargs)
        instance.__init__(*args, **kwargs)
        return instance

    @classmethod
    def from_serializable(cls, object_dict):
        """ core class method to create visible objects from a dictionary """

        key_class = cls._from_visible(cls.STARTS_WITH + 'class' + cls.ENDS_WITH)
        key_module = cls._from_visible(cls.STARTS_WITH + 'module' + cls.ENDS_WITH)

        obj_class = object_dict.pop(key_class)
        obj_module = object_dict.pop(key_module) if key_module in object_dict else None

        obj = cls._from_class(obj_class, obj_module)
        obj.modify_object(object_dict)
        return obj

    @classmethod
    def from_json(cls, json_str):
        obj_dict = json.loads(json_str, object_hook=_decode_dict)
        if isinstance(obj_dict, dict):
            return cls.from_serializable(obj_dict)
        else:
            return [cls.from_serializable(d) for d in obj_dict]

    def to_serializable(self, level=0, all_properties=False):
        d = dict()
        for a in [a for a in dir(self) if self.__class__._is_visible(a)]:
            if a in self._modified_members or self._from_visible(a) in ['Name', 'Class', 'Module'] or all_properties:
                v = getattr(self, a)
                v = v if not hasattr(v, 'to_serializable') else v.to_serializable(level + 1)
                v = v if isinstance(v, (float, int, list, type(None))) else str(v)
                d[self.__class__._from_visible(a)] = v
        return d

    def to_json(self, indent=2, all_properties=False):
        obj_dict = self.to_serializable(all_properties=all_properties)
        # todo better pretty print for DataRange and AttributeList (nested list)
        return json.dumps(obj_dict, indent=indent, sort_keys=True)

    def modify_object(self, property_name, property_value=None):
        """
        api visible method for modifying visible object properties

        :param property_name: property name
        :type property_name: string, list or dict
        :param property_value: property value, must be `None` if property_name is of type `dict`
        :type property_value: various or None
        :return: modified object
        :rtype: unicum.lfojbect.VisibleObject
        """
        if type(property_name) is dict:
            property_value = property_name.values()
            property_name = property_name.keys()

        if isinstance(property_name, str):
            property_name, property_value = [property_name], [property_value]
        assert len(property_name) == len(property_value)

        # convert names into visible
        property_name = self.__class__._to_visible(property_name)

        # loop over properties to set
        for n, v in zip(property_name, property_value):
            self._modify_property(n, v)

        # rebuild object in order to maintain consistency
        self._rebuild_object()
        return self

    def _modify_property(self, property_name, property_value):
        # avoid circles
        if property_value is self:
            raise ValueError, 'Attributes must not be recursively. Not mapping self to %s.' % property_name

        # handle not admissible property_name type
        if not isinstance(property_name, str):
            s = str(property_name), type(property_name), self.__class__.__name__
            msg = 'can not handle %s of type %s as a property in object of type %s' % s
            print(msg)
            # raise TypeError, msg
            return

        # handle not admissible property_name
        if not hasattr(self, property_name) or not self.__class__._is_visible(property_name):
            if property_name:
                s = property_name, self.__class__.__name__, str(self)
                msg = 'property %s in object of type %s not found in %s' % s
                print(msg)
                # raise ValueError, msg
                return

        # check type of property_value
        current_property_value = getattr(self, property_name)
        if not isinstance(property_value, type(current_property_value)):
            property_value = current_property_value.__class__(property_value)

        # finally set new property value
        setattr(self, property_name, property_value)

        # update timestamps and version
        self._ts_update = datetime.datetime.now()
        self._user_update = getpass.getuser() if hasattr(getpass, 'getuser') else 'NoUser'
        self._version += 1
        self._modified_members.append(property_name)

    def _rebuild_object(self):
        """ method to initiate a visible object rebuild """
        pass


# container class of list of objects
class AttributeList(list):
    """ object list class """

    def __reduce__(self):
        return self.__class__, (list(self), self._object_type, self._value_types)

    def __init__(self, iterable=None, object_type=PersistentObject, value_types=(float, int, str, type(None))):
        if not issubclass(object_type, PersistentObject):
            raise TypeError, \
                'Required object type of AttributeList items must be subtype of %s ' % PersistentObject.__name__
        self._object_type = object_type
        self._value_types = value_types
        if iterable is None or len(iterable) is 0:  # iterable is None
            super(AttributeList, self).__init__()
        else:
            # must be or build object_type
            # list of lists -> list of dicts
            if all([type(x) is list for x in iterable]):  # iterable is nested list of object properties
                keys = tuple(iterable.pop(0))  # extract headline of property names
                if not len(keys) == len(set(keys)):
                    raise ValueError, 'Properties in AttributeList must be unique'
                iterable = [self._object_from_serializable(x, keys) for x in iterable]
            elif all([type(x) is dict for x in iterable]):  # iterable is list of dict of object properties
                iterable = [self._object_from_serializable(x) for x in iterable]
            # validate objects
            for x in iterable:
                self._validate(x)
            super(AttributeList, self).__init__(iterable)

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self) + ', %s)' % self._object_type.__name__

    def __str__(self):
        return '[' + ', '.join(repr(x) for x in self) + ']'

    def _object_from_serializable(self, x, keys=None):
        if isinstance(x, list):
            x = dict(zip(keys, x))
        if isinstance(x, dict):
            x = self._object_type.from_serializable(x)
        return x if isinstance(x, self._object_type) else self._object_type(x)

    def _validate(self, x):
        if not isinstance(x, self._object_type):
            raise TypeError, 'All items in this AttributeList must have subtype of %s.' % self._object_type.__name__
        for k, v in x.to_serializable().items():
            if not isinstance(v, self._value_types):
                s = ', '.join([str(t) for t in self._value_types])
                t = type(v)
                raise TypeError, \
                    'All properties of item in this AttributeList must have type ' \
                    'of either one of %s but not %s as seen for property %s' % (s, t, k)

    def __iter__(self):
        return super(AttributeList, self).__iter__()

    def __setitem__(self, key, value):
        value = self._object_from_serializable(value)
        self._validate(value)
        super(AttributeList, self).__setitem__(key, value)

    def __add__(self, iterable):
        iterable = [self._object_from_serializable(value) for value in iterable]
        for value in iterable:
            self._validate(value)
        return self.__class__(super(AttributeList, self).__add__(iterable), self._object_type)

    def __iadd__(self, iterable):
        iterable = [self._object_from_serializable(value) for value in iterable]
        for value in iterable:
            self._validate(value)
        return self.__class__(super(AttributeList, self).__iadd__(iterable), self._object_type)

    def __setslice__(self, i, j, iterable):
        iterable = [self._object_from_serializable(value) for value in iterable]
        for value in iterable:
            self._validate(value)
        super(AttributeList, self).__setslice__(i, j, iterable)

    @classmethod
    def from_serializable(cls, item):
        return cls(item)

    def append(self, value):
        value = self._object_from_serializable(value)
        self._validate(value)
        super(AttributeList, self).append(value)

    def index(self, value, start=None, stop=None):
        value = self._object_from_serializable(value)
        self._validate(value)
        return super(AttributeList, self).index(value, start, stop)

    def insert(self, index, value):
        value = self._object_from_serializable(value)
        self._validate(value)
        super(AttributeList, self).insert(index, value)

    def extend(self, iterable):
        iterable = [self._object_from_serializable(value) for value in iterable]
        for value in iterable:
            self._validate(value)
        super(AttributeList, self).extend(iterable)

    def to_serializable(self, level=0):
        if not self:
            return [['Name', 'Class', 'Module']]
        # list of objects -> list of dicts
        d = [x.to_serializable() for x in self]
        keys = sorted(self.keys())
        ret = [keys]
        for o in d:
            l = list()
            for k in keys:
                v = o.get(k, None)
                v = v if not hasattr(v, 'to_serializable') else v.to_serializable(level + 1)
                v = v if isinstance(v, self._value_types) else str(v)
                # v = v if isinstance(v, (float, int, type(None))) else str(v)
                l.append(v)
            ret.append(l)
        return ret

    def keys(self):
        return sorted(set().union(*[x.to_serializable().keys() for x in self]))

    def values(self):
        keys = self.keys()
        dicts = [x.to_serializable() for x in self]
        return [[o.get(k) for k in keys] for o in dicts]

    def items(self):
        return zip(self.keys(), self.items())



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


import json


def decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
            if type(item) == type(""):
                if not item.isdigit():
                    try:
                        item = float(item)
                    except:
                        item = item
        elif isinstance(item, list):
            item = decode_list(item)
        elif isinstance(item, dict):
            item = decode_dict(item)
        rv.append(item)
    return rv


def decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = decode_list(value)
        elif isinstance(value, dict):
            value = decode_dict(value)
        rv[key] = value
    return rv


class NestedList(list):

    def __init__(self, iterable=[]):
        if iterable:
            super(NestedList, self).__init__([iterable])
        else:
            super(NestedList, self).__init__([])


class NestedListEncoder(json.JSONEncoder):
    @staticmethod
    def _is_nested_list(obj):
        return all(isinstance(o, list) for o in obj) if isinstance(obj, list) else False

    def default(self, obj):
        if isinstance(obj, NestedList):
            return ',\n'.join([json.JSONEncoder().encode(o) for o in obj])
        return json.JSONEncoder.default(self, obj)

    def encode(self, obj):
        if self._is_nested_list(obj):
            return ',\n'.join([json.JSONEncoder().encode(o) for o in obj])
        # Let the base class default method raise the TypeError
        s = super(NestedListEncoder, self).encode(obj)
        return s

    def iterencode(self, obj, _one_shot=True):
        if self._is_nested_list(obj):
            return ',\n'.join([json.JSONEncoder().encode(o) for o in obj])
        # Let the base class default method raise the TypeError
        s = super(NestedListEncoder, self).iterencode(obj, 1)
        return s



# -*- coding: utf-8 -*-

#  unicum
#  ------------
#  Simple object cache and __factory.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/unicum
#  License: APACHE Version 2 License (see LICENSE file)


from datetime import datetime
from json import dumps
from copy import copy, deepcopy
from unittest import TestCase, main, TestLoader, TextTestRunner
from os import getcwd
from unicum import DecoratorFactory, SingletonObject, FactoryObject, ObjectList, \
    LinkedObject, PersistentObject, AttributeList, VisibleObject, VisibleAttributeList
from unicum.datarange import DataRange
from unicum.olddatarange import IndexedList


class TestDummy(object):
    pass


class CacheTest(TestCase):
    def setUp(self):
        self.cache = DecoratorFactory()

        class UnnamedObject(object):
            def __init__(self, *args, **kargs):
                pass

            def __str__(self):
                return self.__class__.__name__

            def to_serializable(self, level=0):
                return str(self)

        class DayCount(UnnamedObject):
            pass

        self.DayCount = DayCount

    def test_NamedObject(self):
        # - test NamedObject -

        @self.cache.by_key
        class NamedObject(object):
            def __init__(self, name='last'):
                self.name = name

        first = NamedObject('First')
        self.cache['Second'] = first
        second = NamedObject('second')
        self.assertEqual(first, second)

        third = NamedObject('Third')
        self.assertNotEqual(first, third)
        self.assertEqual(third, NamedObject('third'))

        self.cache.pop('third')
        self.assertNotEqual(third, NamedObject('third'))

        self.cache['Forth'] = NamedObject('first')
        forth = NamedObject('Forth')
        self.assertEqual(NamedObject('first'), forth)
        self.assertEqual(second, NamedObject('forth'))

        self.cache['Fifth'] = NamedObject('first')
        self.assertEqual(second, NamedObject('fifth'))
        self.cache.pop('FIFTH')
        self.assertNotIn('fifth', self.cache)

        last = NamedObject()
        self.assertNotEqual(last, NamedObject())
        self.assertNotIn('NamedObject', self.cache)

        name_objects = DecoratorFactory(self.cache.cache, NamedObject)
        with self.assertRaises(KeyError):
            name_objects()

    def test_UnnamedObject(self):
        # - test UnnamedObject -

        @self.cache.by_class_name
        class ActAct(self.DayCount):
            pass

        @self.cache.by_class_name
        class Thirty360(self.DayCount):
            pass

        Thirty360()
        self.cache['Act/Act'] = ActAct()
        self.cache['Act_Act'] = ActAct()
        self.assertEqual(self.cache['Act/ACT'], ActAct())
        self.assertEqual(self.cache['ACTACT'], self.cache['ACT/act'])

        day_count = DecoratorFactory(self.cache.cache, self.DayCount, 'ActAct')
        self.assertEqual(day_count(), ActAct())
        self.assertEqual(day_count('actact'), ActAct())

        self.assertEqual(ActAct().to_serializable(), 'ActAct')
        self.assertEqual(self.cache['Act/ACT'].to_serializable(), 'ActAct')


class SingletonTest(TestCase):
    def setUp(self):
        class SingeltonDummy(TestDummy, SingletonObject):
            pass

        self.Constant = SingeltonDummy

    def test_singleton(self):
        c = self.Constant()
        d = self.Constant()
        self.assertEqual(c, d)


class FactoryTest(TestCase):
    def setUp(self):
        class Currency(FactoryObject):
            __factory = dict()

            def __new__(cls, *args, **kwargs):
                if not args:
                    return super(Currency, cls).__new__(cls, cls.__name__)
                else:
                    return super(Currency, cls).__new__(cls, *args)

            def __init__(self, name=None):
                super(Currency, self).__init__(name)
                self._vp_ = self.__class__.__name__

        self.Currency = Currency

        class EUR(Currency):
            @property
            def vp(self):
                return self._vp_

        self.EUR = EUR

        class USD(Currency):
            @property
            def vp(self):
                return self._vp_

        self.USD = USD

        class CurrencyList(ObjectList):
            def __init__(self, iterable=None):
                super(CurrencyList, self).__init__(iterable, Currency)

            def __add__(self, other):
                return CurrencyList(super(CurrencyList, self).__add__(other))

            def __iadd__(self, other):
                return CurrencyList(super(CurrencyList, self).__iadd__(other))

        self.CurrencyList = CurrencyList

        class Interpolation(FactoryObject):
            __factory = dict()

        class FactoryDummy(TestDummy, Interpolation):
            pass

        self.FactoryDummy = FactoryDummy

        class AnotherFactoryDummy(TestDummy, Interpolation):
            pass

        self.AnotherFactoryDummy = AnotherFactoryDummy

        class FactoryDummySubClass(AnotherFactoryDummy):
            pass

        self.FactoryDummySubClass = FactoryDummySubClass

    def test_first_factory(self):
        e = self.FactoryDummy().register()
        f = self.AnotherFactoryDummy().register()
        n = self.FactoryDummySubClass().register()
        g = n.__class__(str(f))
        self.assertTrue(e is not f)
        self.assertTrue(g is f)

    def test_second_factory(self):
        # test FactoryObject
        eur = self.EUR().register()
        self.EUR().register('eur')

        usd = self.USD().register()
        self.assertTrue(eur is not usd)

        neur = self.USD('EUR')
        self.assertTrue(eur is neur)
        self.assertTrue(self.USD('EUR').vp is 'EUR')
        self.assertTrue(self.USD('USD').vp is 'USD')

        fEUR = self.Currency('EUR')
        feur = self.Currency('eur')
        self.assertTrue(type(eur) is type(fEUR))
        self.assertTrue(eur is feur)
        self.assertTrue(fEUR is feur)

        for x, y in zip(self.EUR.values(), self.USD.values()):
            self.assertTrue(x is y)

        for ek, (k, v), ev in zip(self.EUR.keys(), self.USD.items(), self.Currency.values()):
            self.assertTrue(ek is k)
            self.assertTrue(v is ev)

        self.Currency('EUR').remove()
        self.assertTrue(len(self.EUR.items()) is 1)
        self.assertTrue(self.EUR.values()[0] is usd)
        self.assertTrue(usd in self.Currency.values())
        self.assertTrue(eur not in self.Currency.values())

        # NamedObject -> create item by self.__class__(obj_name)
        # NamedObject by SingletonObject
        # NamedObject from FactoryObject

        # Idempotent, Strong or SingletonNamedObject
        # e.g. __factory, singleton subs
        ccy = self.EUR().register()
        self.assertTrue(ccy == ccy.__class__(ccy.to_serializable()))
        self.assertTrue(ccy == ccy.__class__(ccy.__str__()))
        self.assertTrue(ccy == ccy.__class__(ccy))
        self.assertTrue(ccy == self.Currency(ccy))

        # Equivalent, Weak or SimpleNamedObject
        # e.g. BusinessDate, BusinessPeriod
        ccy = self.EUR()
        self.assertTrue(str(ccy) is str(ccy.__class__(ccy.__str__())))
        self.assertTrue(str(ccy) is str(ccy.__class__(ccy.__str__())))

    def test_register(self):
        # test FactoryObject
        eur = self.EUR().register()
        self.assertTrue('EUR' in self.EUR.keys())

        names = 'eur', 'Eur', 'EURO', 'euro'
        self.EUR().register(*names)
        keys = self.EUR.keys()
        for n in names:
            self.assertTrue(n in keys)

    def test_mixed(self):
        usd = self.USD().register()
        const = self.FactoryDummy().register()

        self.assertTrue(usd in self.USD.values())
        self.assertTrue(usd not in self.FactoryDummy.values())

        self.assertTrue(const not in self.USD.values())
        self.assertTrue(const in self.FactoryDummy.values())

    def test_objectList(self):
        eur, usd = self.EUR().register(), self.USD().register()
        l = (eur, usd)
        o = ObjectList(l, self.Currency)
        o = self.CurrencyList(o)
        self.assertTrue(o == o.__class__(l))
        self.assertTrue(o == o.__class__(o))
        self.assertTrue(o == o.__class__(o.to_serializable()))

        for x, y in zip(l, o.to_serializable()):
            self.assertTrue(type(y) is str)
            self.assertTrue(str(x) is y)

        const = self.FactoryDummy().register()
        ol = (lambda x: ObjectList(x, object_type=self.Currency))
        self.assertRaises(TypeError, ol, (eur, const))

        eur.register('eur')
        usd.register('usd')

        o[0] = eur
        self.assertTrue(eur in o)
        o.pop(0)
        self.assertTrue(eur not in o)
        o[0] = 'eur'
        self.assertTrue(eur in o)
        o.pop(0)
        o.insert(0, eur)
        self.assertTrue(eur in o)
        o.pop(0)
        o.insert(0, 'eur')
        self.assertTrue(eur in o)
        o.pop(0)
        o.append(eur)
        self.assertTrue(eur in o)
        o.pop(0)
        o.append('eur')
        self.assertTrue(eur in o)
        o.pop(0)
        o.extend([eur])
        self.assertTrue(eur in o)
        o.pop(0)
        o.extend(['eur'])
        self.assertTrue(eur in o)
        o.pop(0)
        o[0:0] = [eur]
        self.assertTrue(eur in o)
        o.pop(0)
        o[0:0] = ['eur']
        self.assertTrue(eur in o)

        b = o + ['eur']
        self.assertTrue(isinstance(b, self.CurrencyList))
        self.assertTrue(eur in b)

        b = ['eur'] + o
        self.assertTrue(not isinstance(b, ObjectList))
        self.assertTrue(eur in b)

    def test_get_item(self):
        eur, usd = self.EUR().register(), self.USD().register()
        l = (eur, usd)
        o = ObjectList(l, self.Currency)
        o = self.CurrencyList(o)
        self.assertEqual(o['EUR'], eur)
        self.assertTrue(eur in o)
        self.assertEqual(o.index(eur), 0)
        self.assertEqual(o.get(eur), eur)
        self.assertEqual(o.get('EUR'), eur)

class MyLO(LinkedObject):
    def __init__(self):
        super(MyLO, self).__init__()
        self.property = LinkedObject()


class YourLO(LinkedObject):
    def __init__(self):
        super(YourLO, self).__init__()
        self.property = 'It is yours.'


class LinkedTest(TestCase):
    def test_object_link(self):
        # test FactoryObject
        # ------------

        m = MyLO()
        n = MyLO()
        self.assertTrue(m.property is not n.property)
        self.assertTrue(len(LinkedObject._get_links()) is 1)
        self.assertTrue(m.property.__class__.__name__ in LinkedObject._get_links())
        self.assertTrue(n.property.__class__.__name__ in LinkedObject._get_links())

        m.property = n.property
        self.assertTrue(m.property is n.property)
        self.assertTrue(len(LinkedObject._get_links()) is 1)
        self.assertTrue(m.property.__class__.__name__ in LinkedObject._get_links())
        self.assertTrue(n.property.__class__.__name__ in LinkedObject._get_links())

        z = YourLO()
        y = YourLO()
        m.property = n.property = y
        self.assertTrue(m.property is y)
        self.assertTrue(m.property is not z)
        self.assertTrue(m.property is n.property)
        self.assertTrue(len(LinkedObject._get_links()) is 1)
        self.assertTrue(m.property.__class__.__name__ in LinkedObject._get_links())
        self.assertTrue(n.property.__class__.__name__ in LinkedObject._get_links())

        z.update_link()
        self.assertTrue(m.property is not y)
        self.assertTrue(m.property is z)
        self.assertTrue(m.property is n.property)
        self.assertTrue(len(LinkedObject._get_links()) is 1)
        self.assertTrue(m.property.__class__.__name__ in LinkedObject._get_links())
        self.assertTrue(n.property.__class__.__name__ in LinkedObject._get_links())


class MyPO(PersistentObject):
    def __init__(self):
        super(MyPO, self).__init__()
        self._my_property_ = 'It is mine.'


class YourPO(PersistentObject):
    def __init__(self):
        super(YourPO, self).__init__()
        self._your_property_ = 'It is yours.'


class PersistentTest(TestCase):
    def test_obj(self):
        p = MyPO()
        self.assertTrue(type(p) is MyPO)
        self.assertTrue(hasattr(p, '_class_'))

    def test_visible(self):
        for a in ['_class_', 'ObjectName']:
            if MyPO._is_visible(a):
                self.assertEqual(a, MyPO._to_visible(a))
            else:
                self.assertEqual(a, MyPO._from_visible(a))
                # print a, MyPO._is_visible(a), MyPO._to_visible(a), MyPO._from_visible(a)

    def test_obj_to_dict(self):
        d = MyPO().to_serializable(all_properties=True)
        self.assertTrue('Class' in d)
        self.assertTrue(d['MyProperty'] == 'It is mine.')

    def test_dict_to_obj(self):
        e = {'Class': 'PersistentObject'}
        self.assertTrue(type(MyPO.from_serializable(e)) is PersistentObject)
        e = {'Class': 'YourPO', 'Module': __name__}
        self.assertTrue(MyPO.from_serializable(e).to_serializable()['Class'] == 'YourPO')

    def test_json(self):
        e = {'Class': 'PersistentObject'}
        j = dumps(e)
        o = PersistentObject.from_json(j)
        self.assertEqual(type(o), PersistentObject)

        e = {'Class': 'YourPO', 'Module': __name__, 'YourProperty': 'It is mine.'}
        j = dumps(e, indent=2, sort_keys=True)
        o = PersistentObject.from_json(j)
        self.assertTrue(type(o) is YourPO)
        self.assertEqual(o.to_json(indent=2), j)
        self.assertEqual(o.to_serializable(), PersistentObject.from_serializable(e).to_serializable())

    def test_modify_obj(self):
        q = MyPO()
        q.modify_object('MyProperty', 'Hello World.')
        self.assertTrue(q._my_property_ == 'Hello World.')
        d = q.to_serializable()
        self.assertTrue(d['MyProperty'] == 'Hello World.')

        # no direct circle assignment
        self.assertRaises(ValueError, q.modify_object, 'MyProperty', q)

    def test_modify_factory_obj(self):
        class MyFactoryObject(FactoryObject):
            pass

        q = MyPO()
        fo = FactoryObject().register()
        fo.register('MyObj')
        mfo = MyFactoryObject().register()

        q._my_property_ = FactoryObject()
        self.assertTrue(q._my_property_ == fo)
        q.modify_object('MyProperty', 'MyFactoryObject')
        self.assertTrue(q._my_property_ == mfo)
        q.modify_object('MyProperty', 'MyObj')
        self.assertTrue(q._my_property_ == fo)
        q.modify_object('MyProperty', MyFactoryObject())
        self.assertTrue(q._my_property_ == mfo)

    def test_attribute_list(self):
        l = MyPO(), YourPO()
        a = AttributeList(l)

        self.assertTrue(a == a.__class__(l))
        self.assertTrue(a == a.__class__(a))

    def test_attribute_list2(self):
        ol = (lambda x: AttributeList(x, object_type=MyPO))
        l = MyPO(), MyPO()
        self.assertTrue(len(ol(l)) == 2)
        l = YourPO(), MyPO()
        self.assertRaises(TypeError, ol, l)

    def test_attribute_list3(self):
        l = YourPO(), MyPO()
        a = AttributeList(l)
        m = MyPO()

        a[0] = m
        self.assertTrue(m in a)
        a.pop(0)
        self.assertTrue(m not in a)

        a.insert(0, m)
        self.assertTrue(m in a)
        a.pop(0)
        self.assertTrue(m not in a)

        a.append(m)
        self.assertTrue(m in a)
        a.pop(-1)
        self.assertTrue(m not in a)

        a.extend([m])
        self.assertTrue(m in a)
        a.pop(-1)
        self.assertTrue(m not in a)

        a[0:0] = [m]
        self.assertTrue(m in a)
        a.pop(0)
        self.assertTrue(m not in a)

        b = a + [m]
        self.assertTrue(isinstance(b, AttributeList))
        self.assertTrue(m in b)

        b = [m] + a
        self.assertTrue(not isinstance(b, AttributeList))
        self.assertTrue(m in b)


class IndexedListTest(TestCase):
    def setUp(self):
        self.values = range(100)
        self.keys = [str(i) for i in self.values]
        self.list = IndexedList(self.keys, self.values)

    def test_list(self):
        indexed_list = IndexedList(self.list)
        for il, org in zip(indexed_list.keys(), self.keys):
            self.assertEqual(il, org)
        for il, org in zip(indexed_list.keys(), self.values):
            self.assertEqual(indexed_list[il], org)
        for il, org in zip(indexed_list.values(), self.values):
            self.assertEqual(il, org)
        for k, v in indexed_list.items():
            self.assertEqual(k, str(v))

        self.assertEqual(indexed_list[2:50], indexed_list['2':'49'])
        self.assertEqual(indexed_list[2:], indexed_list['2':])
        self.assertEqual(indexed_list[:2], indexed_list[:'1'])

        new = ['A', 'B']
        indexed_list['3':'5'] = new
        self.assertEqual(indexed_list[3], new[0])
        self.assertEqual(indexed_list[4], new[1])
        self.assertEqual(indexed_list[3:5], new)

    def test_set(self):
        indexed_list = IndexedList(self.list)
        self.assertEqual(len(indexed_list), 100)

        indexed_list.append(100)
        self.assertEqual(len(indexed_list), 101)
        self.assertEqual(indexed_list[100], 100)

        indexed_list.insert('50', -1)
        self.assertEqual(len(indexed_list), 102)
        self.assertEqual(indexed_list[50], -1)
        self.assertEqual(indexed_list[49], 49)
        self.assertEqual(indexed_list[51], 50)

        indexed_list.extend(['a'], ['b'])
        self.assertEqual(indexed_list.keys()[-1], 'a')
        self.assertEqual(indexed_list[-1], 'b')

        new_il = IndexedList(['c'], ['d'])
        indexed_list.extend(new_il)
        self.assertEqual(indexed_list.keys()[-1], 'c')
        self.assertEqual(indexed_list[-1], 'd')

        new_il = IndexedList(['c'], ['c'])
        l = (lambda x: indexed_list.extend(x))
        self.assertRaises(KeyError, l, new_il)


class DataRangeTest(TestCase):
    def setUp(self):
        h = '', 'A', 'B', 'C', 'D'
        x = 'X', 1, 2, 3, 4
        y = 'Y', 4, 5, 6, 7
        z = 'Z', 7, 8, 9, 10
        self.datarange = DataRange([h, x, y, z])
        pass

    def test_keys(self):
        self.assertEqual(self.datarange.col_keys(), ['A', 'B', 'C', 'D'])
        self.assertEqual(self.datarange.row_keys(), ['X', 'Y', 'Z'])

    def test_get(self):
        self.assertEqual(self.datarange.get(('X', 'A')), 1)
        self.assertEqual(self.datarange.row('X'), [1, 2, 3, 4])
        self.assertEqual(self.datarange.col('D'), [4, 7, 10])
        # self.assertEqual(self.datarange[0][-1], 4)
        self.assertEqual(self.datarange[0, -1], 4)
        # self.assertEqual(self.datarange['X']['C'], 3)
        self.assertEqual(self.datarange['X', 'C'], 3)

    def test_slice(self):
        self.assertEqual(self.datarange[0:2], self.datarange['X':'Y'])
        self.assertEqual(self.datarange[0:1], [[1, 2, 3, 4]])
        self.assertEqual(self.datarange[(0, 1):(2, 3)], [[2, 3], [5, 6]])
        # self.assertEqual([l[1:3] for l in self.datarange[0:2]], [[2, 3], [5, 6]])
        # self.assertEqual([l[1:3] for l in self.datarange['X':'Y']], [[2, 3], [5, 6]])
        # self.assertEqual([l['B':'C'] for l in self.datarange['X':'Y']], [[2, 3], [5, 6]])
        self.assertEqual(self.datarange[('X', 'B'):('Y', 'C')], [[2, 3], [5, 6]])
        self.assertEqual(self.datarange[(None, 'B'):('Y', 'C')], [[2, 3], [5, 6]])
        self.assertEqual(self.datarange[(None, 'B'):('Y', None)], [[2, 3, 4], [5, 6, 7]])

    def test_flatten(self):
        self.assertEqual(DataRange(self.datarange.to_serializable()), self.datarange)
        s = "DataRange([[None, 'A', 'B', 'C', 'D'], ['X', 1, 2, 3, 4], ['Y', 4, 5, 6, 7], ['Z', 7, 8, 9, 10]])"
        self.assertEqual(str(self.datarange), s)

    def test_set(self):
        # self.datarange['X']['C'] = 4
        # self.assertEqual(self.datarange['X']['C'], 4)
        self.datarange['X', 'C'] = 3
        l = self.datarange.__setitem__
        self.assertRaises(KeyError, l, ('U', 'C'), 4)
        self.assertRaises(KeyError, l, ('X', 'X'), 4)

    def test_append(self):
        self.datarange.append('W', range(4))
        l = lambda: self.datarange.row_append('W', range(4))
        self.assertRaises(KeyError, l)
        self.assertEqual(len(self.datarange), 4)
        self.assertEqual(self.datarange.row_keys()[-1], 'W')
        self.assertEqual(self.datarange.row(-1), range(4))

        self.datarange.row_append('U', range(2, 6))
        self.assertEqual(self.datarange.row('U'), range(2, 6))
        self.assertEqual(self.datarange.row(4), range(2, 6))
        self.assertEqual(self.datarange.row(-1), range(2, 6))

        self.datarange.col_append('T', range(5))
        self.assertEqual(self.datarange.col('T'), range(5))
        self.assertEqual(self.datarange.col(4), range(5))
        self.assertEqual(self.datarange.col(-1), range(5))

    def test_copy(self):
        self.assertEqual(self.datarange, copy(self.datarange))
        self.assertEqual(type(self.datarange), type(copy(self.datarange)))
        self.assertEqual(self.datarange, deepcopy(self.datarange))
        self.assertEqual(type(self.datarange), type(deepcopy(self.datarange)))

    def test_transpose(self):
        self.assertEqual(type(list(self.datarange)), list)
        l = [self.datarange.row(r) for r in self.datarange.row_keys()]
        self.assertEqual(self.datarange.item_list, l)
        self.assertEqual(type(self.datarange), type(self.datarange.transpose()))

    def test_pickle(self):
        try:
            import dill as pickle
        except ImportError:
            pass
        else:
            dr = DataRange()
            p = pickle.dumps(dr)
            d = pickle.loads(p)
            self.assertEqual(type(d), DataRange)
            self.assertEqual(dr.to_serializable(), d.to_serializable())

            self.assertEqual(type(self.datarange), DataRange)
            p = pickle.dumps(self.datarange)
            d = pickle.loads(p)
            self.assertEqual(self.datarange, d)


class MyVO(VisibleObject):
    def __init__(self, *args):
        super(MyVO, self).__init__(*args)
        self._none_prop_ = None
        self._str_prop_ = str('my str')
        self._int_prop_ = int(100)
        self._flt_prop_ = float(99.01)
        self._obj_prop_ = VisibleObject('YourVisibleObject')
        self._obj_list_prop_ = ObjectList()
        self._attr_list_prop_ = AttributeList()
        self._data_range_prop_ = DataRange()

    @property
    def none(self):
        return self._none_prop_

    @property
    def str(self):
        return self._str_prop_

    @property
    def int(self):
        return self._int_prop_

    @property
    def flt(self):
        return self._flt_prop_

    @property
    def obj(self):
        return self._obj_prop_

    @property
    def obj_list(self):
        return self._obj_list_prop_

    @property
    def attr_list(self):
        return self._attr_list_prop_

    @property
    def datarange(self):
        return self._data_range_prop_


class VisibleTest(TestCase):
    def test_factory(self):
        obj = VisibleObject('MyHello')
        obj.register()
        obj.register('Hello')
        self.assertEqual(obj, VisibleObject('Hello'))

        o = MyVO._from_class('MyVO', __name__, 'MyTrueVO').register()
        self.assertEqual(o, VisibleObject('MyTrueVO'))

        s = obj.to_json().replace('MyHello', 'NewHello')
        so = VisibleObject.from_json(s).register()
        self.assertEqual(so, VisibleObject('NewHello'))

    def test_link(self):
        obj = MyVO('My')
        first = VisibleObject('FirstProperty')
        obj.modify_object('ObjProp', first)

        second = VisibleObject('FirstProperty')
        self.assertEqual(obj.obj, first)
        self.assertNotEqual(obj.obj, second)

        second.update_link()
        self.assertNotEqual(obj.obj, first)
        self.assertEqual(obj.obj, second)

    def test_persistence(self):
        obj = MyVO('My')
        dic = obj.to_serializable(all_properties=True)
        # test None
        self.assertEqual(dic['NoneProp'], obj.none)
        # test str
        self.assertEqual(dic['StrProp'], obj.str)
        # test int
        self.assertEqual(dic['IntProp'], obj.int)
        # test flt
        self.assertEqual(dic['FltProp'], obj.flt)
        # test FactoryObject
        self.assertEqual(dic['ObjProp'], obj.obj.to_serializable(1))
        # test ObjectList
        self.assertEqual(dic['ObjListProp'], obj.obj_list.to_serializable(1))
        # test AttributeList
        self.assertEqual(dic['AttrListProp'], obj.attr_list.to_serializable(1))
        # test DataRange
        self.assertEqual(dic['DataRangeProp'], obj.datarange.to_serializable(1))
        for k, v in obj.to_serializable().items():
            # print k.ljust(16), str(type(v)).ljust(20), v
            self.assertTrue(isinstance(v, (int, float, str, type(None), list)))
        self.assertTrue(obj.to_json())

    def test_objlist(self):
        names = ['obj' + str(i) for i in range(10)]
        l = ObjectList([MyVO(n) for n in names], object_type=VisibleObject)
        self.assertEqual(l.to_serializable(1), names)
        o = l[0]
        self.assertTrue(o in l)

    def test_attrlist(self):
        names = ['obj' + str(i) for i in range(10)]
        l = [VisibleObject(n) for n in names]
        self.assertTrue(VisibleAttributeList(l))

        l = [object() for n in names]
        self.assertRaises(TypeError, VisibleAttributeList, l)

        l = [MyVO(n).modify_object('DataRangeProp', DataRange()) for n in names]
        self.assertRaises(TypeError, VisibleAttributeList, l)

        # accept Attributes with more than depth 2 -> adding list to value typese
        t = int, float, str, type(None), VisibleObject, list
        a = VisibleAttributeList([o.to_serializable() for o in l], value_types=t)
        for r in a:
            self.assertTrue(isinstance(r, VisibleObject))
        for aa in a:
            aa.register()
        self.assertEqual(a[0], VisibleObject(a[0].to_serializable(1)))

        for r in a.to_serializable():
            for c in r:
                self.assertTrue(dumps(c))  # JSON to_serializable
                self.assertTrue(isinstance(c, (int, float, str, type(None), list)))

        obj = MyVO('My')
        obj.modify_object('ObjListProp', l)
        obj.modify_object('AttrListProp', a)
        self.assertTrue(obj.to_json())

    def test_from_class(self):
        VisibleObject('ME').register()
        o = VisibleObject._from_class('VisibleObject', 'unicum', 'ME')
        self.assertEqual(o, VisibleObject('ME'))

if __name__ == '__main__':
    import sys

    start_time = datetime.now()

    print('')
    print('======================================================================')
    print('')
    print('run %s' % __file__)
    print('in %s' % getcwd())
    print('started  at %s' % str(start_time))
    print('')
    print('----------------------------------------------------------------------')
    print('')

    suite = TestLoader().loadTestsFromModule(__import__("__main__"))
    testrunner = TextTestRunner(stream=sys.stdout, descriptions=2, verbosity=2)
    testrunner.run(suite)

    print('')
    print('======================================================================')
    print('')
    print('ran %s' % __file__)
    print('in %s' % getcwd())
    print('started  at %s' % str(start_time))
    print('finished at %s' % str(datetime.now()))
    print('')
    print('----------------------------------------------------------------------')
    print('')



from singletonobject import *
from factoryobject import *
from linkedobject import *
from persistentobject import *
from datarange import *
from visibleobject import *
from genericfactory import *
from cache import *
from cachedobject import *
